import logging
from time import sleep
from openhab_pythonrule_engine.rule_engine import RuleEngine



logging.basicConfig(format='%(asctime)s %(name)-20s: %(levelname)-8s %(message)s', level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
logging.getLogger('sseclient').disabled = True
logging.getLogger('urllib3.connectionpool').disabled = True


rule_engine = RuleEngine.start_singleton("http://192.168.1.27:8080/", "C:\\workspace\\\\openhab_rules", "grro", "Stabilo33!")

sleep(390000)


import logging
from openhab_pythonrule_engine.item_registry import ItemRegistry
from openhab_pythonrule_engine.rule_engine import RuleEngine
from openhab_pythonrule_engine.trigger import CronTrigger, ItemReceivedCommandTrigger, ItemChangedTrigger, RuleLoadedTrigger



def when(target: str):
    """
    Examples:
        .. code-block::
            @when("Time cron 55 55 5 * * ?")
            @when("Item gMotion_Sensors changed")
            @when("Rule loaded")
    Args:
        target (string): the `rules DSL-like formatted trigger expression <https://www.openhab.org/docs/configuration/rules-dsl.html#rule-triggers>`_
            to parse
    """

    target = target.strip()


    if RuleEngine.instance() is None:
        logging.warning("Environment has not been set. Ignoring decorator '@when(" + target + "')")

        def decorated_method(function):
            return function
        return decorated_method

    elif target.lower().startswith("item") and target.lower().endswith(" received command"):
        itemname_operation_pair = target[len("item"):].strip()
        itemname = itemname_operation_pair[:itemname_operation_pair.index(" ")].strip()

        if ItemRegistry.instance().has_item(itemname):
            operation = itemname_operation_pair[itemname_operation_pair.index(" "):].strip()

            def decorated_method(function):
                trigger = ItemReceivedCommandTrigger(itemname, target, function)
                RuleEngine.instance().add_trigger(trigger)
                return function

            return decorated_method
        else:
            logging.warning("item " + itemname + " does not exist (trigger " + target + ")")

    elif target.lower().startswith("item"):
        itemname_operation_pair = target[len("item"):].strip()
        itemname = itemname_operation_pair[:itemname_operation_pair.index(" ")].strip()

        if ItemRegistry.instance().has_item(itemname):
            operation = itemname_operation_pair[itemname_operation_pair.index(" "):].strip()

            def decorated_method(function):
                trigger = ItemChangedTrigger(itemname, operation, target, function)
                RuleEngine.instance().add_trigger(trigger)
                return function

            return decorated_method
        else:
            logging.warning("item " + itemname + " does not exist (trigger " + target + ")")

    elif target.lower().strip() == "rule loaded":

        def decorated_method(function):
            trigger = RuleLoadedTrigger(target, function)
            RuleEngine.instance().add_trigger(trigger)
            return function

        return decorated_method

    elif target.lower().startswith("time cron"):
        cron = target[len("time cron"):].strip()

        def decorated_method(function):
            trigger = CronTrigger(cron, target, function)
            RuleEngine.instance().add_trigger(trigger)
            return function

        return decorated_method

    else:
        logging.warning("unsupported expression " + target + " ignoring it")
        def decorated_method(function):
            return function
        return decorated_method



import time


class Cache:

    def __init__(self):
        self.cache = {}

    def clear(self):
        self.cache.clear()

    def read_entry(self,name:str, max_ttl_sec: int = 60):
        if name in self.cache.keys():
            value_time_pair = self.cache[name]
            if (value_time_pair[1] + max_ttl_sec) > time.time():
                return value_time_pair[0]
            else:
                self.cache.pop(name)
        return None

    def add_enry(self, name, value):
        self.cache[name] = (value, time.time())



from openhab_pythonrule_engine.rule_engine_webthing import run_server
from openhab_pythonrule_engine.app import App
from openhab_pythonrule_engine.rule_engine import RuleEngine
from string import Template



PACKAGENAME = 'openhab_pythonrule_engine'
ENTRY_POINT = "pyrule"
DESCRIPTION = "Openhab python rule engine"



UNIT_TEMPLATE = Template('''
[Unit]
Description=$packagename
After=network-online.target
Wants=network-online.target

[Service]
Type=simple
ExecStart=$entrypoint --command listen --port $port --openhab_uri $openhab_uri --python_rule_directory $python_rule_directory --user $user --pwd $pwd    
SyslogIdentifier=$packagename
StandardOutput=syslog
StandardError=syslog
Restart=always
RestartSec=3

[Install]
WantedBy=multi-user.target
''')




class InternetApp(App):

    def do_add_argument(self, parser):
        parser.add_argument('--openhab_uri', metavar='openhab_uri', required=False, type=str, default="http://localhost:8080", help='the openhab uri such as http://localhost:8080')
        parser.add_argument('--python_rule_directory', metavar='python_rule_directory', required=False, type=str, default="/etc/openhab/automation/rules/python", help='the python_rule_directory such as /etc/openhab/automation/rules/python')
        parser.add_argument('--user', metavar='user', required=True, type=str, help='the user name')
        parser.add_argument('--pwd', metavar='pwd', required=True, type=str, help='the password')

    def do_additional_listen_example_params(self):
        return "--openhab_uri http://localhost:8080 --python_rule_directory /etc/openhab/automation/rules/python --user me --pwd secret"

    def do_process_command(self, command:str, port: int, verbose: bool, args) -> bool:
        if command == 'listen' and (args.openhab_uri is not None) \
                               and (args.python_rule_directory is not None) \
                               and (args.user is not None) \
                               and (args.pwd is not None):
            rule_engine = RuleEngine.start_singleton(args.openhab_uri, args.python_rule_directory, args.user, args.pwd)
            run_server(port, self.description, rule_engine)
            return True
        elif args.command == 'register' and (args.volumio_base_uri is not None):
            print("register " + self.packagename + " on port " + str(args.port))
            unit = UNIT_TEMPLATE.substitute(packagename=self.packagename,
                                            entrypoint=self.entrypoint,
                                            port=port,
                                            openhab_uri=args.openhab_uri,
                                            python_rule_directory=args.python_rule_directory,
                                            user=args.user,
                                            pwd=args.pwd,
                                            verbose=verbose)
            self.unit.register(port, unit)
            return True
        else:
            return False

def main():
    InternetApp(PACKAGENAME, ENTRY_POINT, DESCRIPTION).handle_command()


if __name__ == '__main__':
    main()





import logging
import requests
from dateutil import parser
from datetime import datetime
from dataclasses import dataclass
from requests.auth import HTTPBasicAuth
from typing import Optional, List, Dict, Any
from openhab_pythonrule_engine.cache import Cache


logging = logging.getLogger(__name__)


@dataclass
class Item:
    item_name: str
    read_only: bool
    group_names: List[str]
    value: Any

    def serialize(self, value) -> Optional[str]:
        pass

    def get_state(self):
        pass

    def get_state_as_text(self) -> str:
        pass

    def get_state_as_boolean(self) -> bool:
        pass

    def get_state_as_numeric(self) -> float:
        pass

    def get_state_as_datetime(self) -> datetime:
        text = self.get_state_as_text()
        return datetime.fromtimestamp(parser.parse(text).timestamp())


@dataclass
class TextItem(Item):
    value: str

    def get_state(self):
        return self.get_state_as_text()

    def get_state_as_text(self) -> str:
        return self.value

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return True if (self.get_state_as_text() .lower() in ["true", "on"]) else False

    def get_state_as_numeric(self) -> float:
        if self.value is None:
            return -1
        else:
            return float(self.value)

    def serialize(self, value_to_serialize) -> Optional[str]:
        if value_to_serialize is None:
            return None
        elif type(value_to_serialize) == bool:
            return "ON" if value_to_serialize else "OFF"
        elif type(value_to_serialize) == datetime:
            return value_to_serialize.strftime('%Y-%m-%dT%H:%M:%S')
        elif type(value_to_serialize) == int:
            return str(float(value_to_serialize))
        else:
            return str(value_to_serialize)


@dataclass
class NumericItem(Item):
    item_name: str
    read_only: bool
    value: float

    def get_state(self):
        return self.get_state_as_numeric()

    def get_state_as_numeric(self) -> float:
        return self.value

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return self.value != 0

    def get_state_as_text(self) -> str:
        if self.value is None:
            return ""
        return str(self.value)

    def serialize(self, value_to_serialize) -> Optional[str]:
        if value_to_serialize is None:
            return None
        else:
            if type(value_to_serialize) == bool:
                return "1" if value_to_serialize else "0"
            elif type(value_to_serialize) == int:
                return str(float(value_to_serialize))
            else:
                return str(value_to_serialize)


@dataclass
class BooleanItem(Item):
    item_name: str
    read_only: bool
    value: bool

    def get_state(self):
        return self.get_state_as_boolean()

    def get_state_as_boolean(self) -> bool:
        if self.value is None:
            return False
        else:
            return self.value

    def get_state_as_numeric(self) -> float:
        if self.value is None:
            return 0
        else:
            return 1 if self.value else 0

    def get_state_as_text(self) -> str:
        if self.value is None:
            return str(False)
        else:
            return str(self.value)

    def serialize(self, value_to_serialize) -> Optional[str]:
        if type(value_to_serialize) == bool:
            return "ON" if value_to_serialize else "OFF"
        elif type(value_to_serialize) == float:
            return "ON" if value_to_serialize == 1.0 else "OFF"
        elif type(value_to_serialize) == int:
            return "ON" if value_to_serialize == 1 else "OFF"
        else:
            return "ON" if (str(value_to_serialize).lower() in ["true", "on"]) else "OFF"


def to_item(data) -> Optional[Item]:
    try:
        if 'stateDescription' in data.keys():
            read_only = data['stateDescription']['readOnly']
        else:
            read_only = False
        if 'groupNames' in data.keys():
            group_names = data['groupNames']
        else:
            group_names = []
        state = data['state']
        if data['type'] == 'Number':
            item = NumericItem(data['name'], read_only, group_names, None if (state == 'NULL' or state == 'UNDEF') else float(state))
        elif data['type'] == 'Switch':
            item = BooleanItem(data['name'], read_only, group_names, None if state == 'NULL' else state == 'ON')
        else:
            item = TextItem(data['name'], read_only, group_names, None if (state == 'NULL' or state == 'UNDEF') else state)
        return item
    except Exception as e:
        logging.warning("error occurred mapping " + str(data) + " to item", e)
        return None




class ItemRegistry:
    __instance = None

    @staticmethod
    def new_singleton(openhab_uri: str, user: str, pwd: str):
        item_registry = ItemRegistry(openhab_uri, user, pwd)
        ItemRegistry.__instance = item_registry
        return item_registry

    @staticmethod
    def instance():
        return ItemRegistry.__instance

    def __init__(self, openhab_uri: str, user: str, pwd: str):
        self.cache = Cache()
        self.credentials = HTTPBasicAuth(user, pwd)
        if openhab_uri.endswith("/"):
            self.openhab_uri = openhab_uri
        else:
            self.openhab_uri = openhab_uri + "/"

    def on_event(self, event):
        if event.get("type", "") == "ThingUpdatedEvent":
            logging.info("config change. reset cache")
            self.cache.clear()

    def get_items(self, use_cache: bool = False) -> Dict[str, Item]:
        items = self.cache.read_entry("items", 24 * 60 * 60)
        if items is not None:
            return items
        else:
            uri = self.openhab_uri+ "rest/items"
            try:
                response = requests.get(uri, headers={"Accept": "application/json"}, auth = self.credentials)
                if response.status_code == 200:
                    items = {}
                    for entry in response.json():
                        item = to_item(entry)
                        if item is not None:
                            items[item.item_name] = item
                    self.cache.add_enry("items", items)
                    return items
                elif response.status_code == 404:
                    raise Exception("item " +   uri + " not exists " + response.text)
                else:
                    raise Exception("could not read item state " +   uri +  " got error " + response.text)
            except Exception as e:
                logging.warning("error occurred by calling " + uri, e)

    def get_item(self, item_name: str) -> Optional[Item]:
        uri = self.openhab_uri+ "rest/items/" + item_name
        try:
            response = requests.get(uri, headers={"Accept": "application/json"}, auth = self.credentials)
            if response.status_code == 200:
                data = response.json()
                return to_item(data)
            elif response.status_code == 404:
                raise Exception("item " +   uri + " not exists " + response.text)
            else:
                raise Exception("could not read item state " +   uri +  " got error " + response.text)
        except Exception as e:
            logging.warning("error occurred by calling " + uri, e)

    def has_item(self, item_name: str) -> bool:
        if item_name is None:
            return False
        else:
            return self.get_item(item_name) != None

    def get_group_membernames(self, group_name) -> List[str]:
        return [item.item_name for item in self.get_items().values() if group_name in item.group_names]

    def set_item_state(self, item_name: str, value: str):
        uri = self.openhab_uri+ "rest/items/" + item_name
        try:
            response = requests.post(uri, data=value, headers={"Content-type": "text/plain"}, auth = self.credentials)
            if response.status_code == 200:
                return
            elif response.status_code == 404:
                raise Exception("item " +   uri + " not exists " + response.text)
            else:
                raise Exception("could not update item state " +   uri +  " got error " + response.text)
        except Exception as e:
            logging.warning("error occurred by performing put on " + uri, e)

    def get_item_metadata(self, item_name: str) -> Optional[Item]:
        items_meta_data = self.get_items(use_cache=True)
        for name in items_meta_data.keys():
            if item_name == name:
                return items_meta_data[item_name]
        return None

    def get_state(self, item_name: str, dflt):
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state()

    def get_state_as_numeric(self, item_name: str, dflt: float=-1) -> float:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_numeric()

    def get_state_as_boolean(self, item_name: str, dflt: bool=False) -> bool:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_boolean()

    def get_state_as_text(self, item_name: str, dflt: str="") -> str:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return dflt
        else:
            return state.get_state_as_text()

    def get_state_as_datetime(self, item_name: str, datetime_string: str="1970-01-01") -> datetime:
        state = self.get_item(item_name)
        if state is None or state.value is None:
            return datetime.fromtimestamp(parser.parse(datetime_string).timestamp())
        else:
            return state.get_state_as_datetime()


    def set_state(self, item_name: str, new_state) -> bool:
        if new_state is None:
            logging.warning("try to set " + item_name + " = None. ignoring it")
        else:
            item_metadata = self.get_item_metadata(item_name)
            if item_metadata is None:
                raise Exception("item " + item_name + " not exists")
            else:
                old_state = self.get_state(item_name, None)
                serialized_old_sate = item_metadata.serialize(old_state)
                serialized_new_state = item_metadata.serialize(new_state)
                if serialized_old_sate != serialized_new_state:
                    try:
                        self.set_item_state(item_name, serialized_new_state)
                        logging.debug("set " + item_name + " = " + serialized_new_state)
                        return True
                    except Exception as e:
                        logging.warning("could not set " + item_name + " = " + serialized_new_state, e)
        return False



import logging
from abc import ABC
from datetime import datetime
from dataclasses import dataclass
from typing import Optional
from openhab_pythonrule_engine.invoke import Invoker
from openhab_pythonrule_engine.item_registry import ItemRegistry
from openhab_pythonrule_engine.eventbus_consumer import ItemEvent



logging = logging.getLogger(__name__)



class Trigger(ABC):

    def __init__(self, expression: str, func):
        self.expression = expression
        self.func = func
        self.invoker = Invoker.create(func)
        self.last_executions = []
        self.last_errors = {}
        self.listeners = set()

    def add_listener(self, listener):
        self.listeners.add(listener)

    def is_valid(self) -> bool:
        return self.invoker is not None

    def invoke(self, item_registry: ItemRegistry):
        logging.debug("executing rule " + self.invoker.name + " (triggerred by '" + self.expression + "')")
        if len(self.last_executions) > 20:
            self.last_executions.pop(0)
        try:
            self.invoker.invoke(item_registry)
            self.last_executions.append(Execution(self, datetime.now(), None))
        except Exception as e:
            self.last_executions.append(Execution(self, datetime.now(), e))
            logging.warning("Error occurred by invoking " + self.name, e)
        for listener in self.listeners:
            try:
                listener(self)
            except Exception as e:
                logging.warning("error occurred by calling listener", e)

    @property
    def module(self) -> str:
        return self.func.__module__

    @property
    def name(self) -> str:
        return self.func.__name__

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __lt__(self, other):
        return self.__str__() < other.__str__()

    def __str__(self):
        return self.expression

    def __repr__(self):
        return self.__str__()


class ManualTrigger(Trigger):

    def __init__(self, func):
        super().__init__("manual", func)


class CronTrigger(Trigger):

    def __init__(self, cron: str, expression: str, func):
        self.cron = cron
        super().__init__(expression, func)


class RuleLoadedTrigger(Trigger):

    def __init__(self, expression: str, func):
        super().__init__(expression, func)


class ItemTrigger(Trigger):

    def __init__(self, expression: str, func):
        super().__init__(expression, func)

    def matches(self, item_event: ItemEvent) -> bool:
        return False


class ItemReceivedCommandTrigger(ItemTrigger):

    def __init__(self, item_name: str, expression: str, func):
        self.item_name = item_name
        super().__init__(expression, func)

    def matches(self, item_event: ItemEvent) -> bool:
        return item_event.item_name == self.item_name and item_event.operation == 'command'


class ItemChangedTrigger(ItemTrigger):

    def __init__(self, item_name: str, operation: str, expression: str, func):
        self.item_name = item_name
        self.operation = operation
        super().__init__(expression, func)

    def matches(self, item_event: ItemEvent) -> bool:
        return item_event.item_name == self.item_name and item_event.operation == 'statechanged'




@dataclass
class Execution:
    trigger: Trigger
    datetime: datetime
    error: Optional[Exception] = None

    def __lt__(self, other):
        return self.datetime < other.datetime

    def __eq__(self, other):
        return self.datetime == other.datetime

    def __str__(self):
        text = self.datetime.strftime("%Y-%m-%d-T%H:%M:%S") + " " + str(self.trigger)
        if self.error is not None:
            text += "  (Error: " + str(self.error) + ")"
        return text

    def __repr__(self):
        return self.__str__()


class TriggerRegistry:

    def __init__(self):
        self.triggers_by_module = {}

    def register(self, trigger: Trigger):
        triggers = self.triggers_by_module.get(trigger.module, set())
        triggers.add(trigger)
        self.triggers_by_module[trigger.module] = triggers

    def deregister(self, module):
        if module in list(self.triggers_by_module.keys()):
            logging.info("removing all " + str(len(self.triggers_by_module[module])) + " triggers of '" + module + "'")
            del self.triggers_by_module[module]

    def get_triggers_by_type(self, type):
        filterred_triggers = []
        for triggers in self.triggers_by_module.values():
            for trigger in triggers:
                if isinstance(trigger, type):
                    filterred_triggers.append(trigger)
        return filterred_triggers




from webthing import (Value, Property, Thing, MultipleThings, WebThingServer)
import tornado.ioloop
import logging
from openhab_pythonrule_engine.rule_engine import RuleEngine, Rule


class RuleThing(Thing):

    def __init__(self, description: str, rule: Rule):
        Thing.__init__(
            self,
            'urn:dev:ops:pythonrule-1',
            'python_rule',
            [],
            description
        )

        self.rule = rule
        rule.add_listener(self.on_rule_executed)

        self.name = Value(rule.module + "." + rule.name)
        self.add_property(
            Property(self,
                     'name',
                     self.name,
                     metadata={
                         'title': 'name',
                         'type': 'string',
                         'description': 'the rule name',
                         'readOnly': True
                     }))

        self.triggers = Value(", " .join([trigger.expression for trigger in rule.triggers]))
        self.add_property(
            Property(self,
                     'triggers',
                     self.triggers,
                     metadata={
                         'title': 'Last assigned triggers',
                         'type': 'string',
                         'description': 'The comma-separated trigger names',
                         'readOnly': True,
                     }))

        self.execute = Value(False, self.__execute)
        self.add_property(
            Property(self,
                     'execute',
                     self.name,
                     metadata={
                         '@type': 'BooleanProperty',
                         'title': 'execute',
                         'type': 'boolean',
                         'description': 'triggers the rule manually',
                         'readOnly': False
                     }))

        self.history = Value("\r\n" .join([str(execution) for execution in rule.last_executions]))
        self.add_property(
            Property(self,
                     'history',
                     self.history,
                     metadata={
                         'title': 'The newest trigger executions',
                         'type': 'string',
                         'description': 'The linebreak-separated newest trigger executions',
                         'readOnly': True,
                     }))


        self.last_execution_date = Value("" if rule.last_execution_date is None else rule.last_execution_date.isoformat())
        self.add_property(
        Property(self,
                 'last_execution_date',
                 self.last_execution_date,
                 metadata={
                     'title': 'Last execution date of the rule',
                     'type': 'string',
                     'description': 'The date of the last execution (iso 8601 string)',
                     'readOnly': True,
                 }))

        self.last_trigger = Value("" if rule.last_trigger is None else rule.last_trigger.expression)
        self.add_property(
            Property(self,
                     'last_executed_trigger',
                     self.last_trigger,
                     metadata={
                         'title': 'Last executed trigger of this rule',
                         'type': 'string',
                         'description': 'The trigger name',
                         'readOnly': True,
                     }))

        self.ioloop = tornado.ioloop.IOLoop.current()

    def __execute(self, do_execute: bool):
        if do_execute:
            self.rule.execute()
            self.execute.notify_of_external_update(False)

    def on_rule_executed(self, rule):
        self.ioloop.add_callback(self.__sync_props)

    def __sync_props(self):
        self.last_execution_date.notify_of_external_update("" if self.rule.last_execution_date is None else self.rule.last_execution_date.isoformat())
        self.last_trigger.notify_of_external_update("" if self.rule.last_trigger is None else self.rule.last_trigger.expression)
        self.triggers.notify_of_external_update(", " .join([trigger.expression for trigger in self.rule.triggers]))
        self.history.notify_of_external_update("\r\n" .join([str(execution) for execution in self.rule.last_executions]))


def run_server(port: int, description: str, rule_engine: RuleEngine):
    rule_webthings = [RuleThing(description, rule) for rule in rule_engine.rules]
    server = WebThingServer(MultipleThings(rule_webthings, 'rule'), port=port, disable_host_validation=True)

    try:
        # start webthing server
        logging.info('starting the server listing on ' + str(port))
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping the server')
        server.stop()
        logging.info('done')




import logging
import os
import sys
import importlib
import pycron
from time import sleep
from threading import Thread
from datetime import datetime
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from typing import List, Set, Optional
from openhab_pythonrule_engine.item_registry import ItemRegistry
from openhab_pythonrule_engine.trigger import TriggerRegistry, CronTrigger, ItemTrigger, RuleLoadedTrigger, ManualTrigger, Trigger, Execution
from openhab_pythonrule_engine.eventbus_consumer import EventConsumer, parse_item_event


logging = logging.getLogger(__name__)



class FileSystemListener(FileSystemEventHandler):

    @staticmethod
    def start(rule_engine):
        dir = rule_engine.python_rule_directory
        logging.info("observing rules directory " + dir)
        observer = Observer()
        observer.schedule(FileSystemListener(rule_engine), dir, recursive=False)
        observer.start()
        return observer

    def __init__(self, rule_engine):
        self.rule_engine = rule_engine

    def on_moved(self, event):
        self.rule_engine.unload_module(self.filename(event.src_path))
        self.rule_engine.load_module(self.filename(event.dest_path))

    def on_deleted(self, event):
        self.rule_engine.unload_module(self.filename(event.src_path))

    def on_created(self, event):
        self.rule_engine.load_module(self.filename(event.src_path))

    def on_modified(self, event):
        self.rule_engine.load_module(self.filename(event.src_path))

    def filename(self, path):
        path = path.replace("\\", "/")
        return path[path.rindex("/")+1:]


class CronScheduler:

    def __init__(self):
        self.cron_trigger_by_module = {}
        self.__is_running = True
        self.thread = Thread(target=self.__process, daemon=True)

    def add_job(self, cron_trigger: CronTrigger):
        cron_triggers = self.cron_trigger_by_module.get(cron_trigger.module, set())
        cron_triggers.add(cron_trigger)
        self.cron_trigger_by_module[cron_trigger.module] = cron_triggers

    def remove_jobs(self, module: str):
        if module in self.cron_trigger_by_module.keys():
            logging.info("removing all " + str(len(self.cron_trigger_by_module[module])) + " crons of '" + module + "'")
            del self.cron_trigger_by_module[module]

    def __process(self):
        logging.info("cron scheduler started")
        while self.__is_running:
            for cron_triggers in list(self.cron_trigger_by_module.values()):
                for cron_trigger in list(cron_triggers):
                    if pycron.is_now(cron_trigger.cron):
                        try:
                            cron_trigger.invoke(ItemRegistry.instance())
                        except Exception as e:
                            logging.warning("Error occurred by executing rule " + cron_trigger.name, e)
            sleep(60)  # minimum 60 sec!

    def start(self):
        self.thread.start()

    def stop(self):
        self.__is_running = False
        Thread.join(self.thread)


class Rule:

    def __init__(self, func):
        self.func = func
        self.__triggers = []
        self.__listeners = set()

    @property
    def module(self) -> str:
        return self.func.__module__

    @property
    def name(self) -> str:
        return self.func.__name__

    @property
    def triggers(self) -> List[Trigger]:
        return self.__triggers

    @property
    def last_executions(self) -> List[Execution]:
        executions: list[Execution] = []
        for trigger in self.__triggers:
            for execution in trigger.last_executions:
                executions.append(execution)
        executions.sort(reverse=True)
        return executions

    @property
    def last_execution_date(self) -> Optional[datetime]:
        execution = self.__newest_execution()
        if execution is None:
            return None
        else:
            return execution.datetime

    @property
    def last_trigger(self) -> Optional[Trigger]:
        execution = self.__newest_execution()
        if execution is None:
            return None
        else:
            return execution.trigger

    def __newest_execution(self) -> Optional[Execution]:
        last_execution = None
        for trigger in self.__triggers:
            for execution in trigger.last_executions:
                if last_execution is None or execution.datetime > last_execution.datetime:
                    last_execution = execution
        return last_execution

    def add_listener(self, listener):
        self.__listeners.add(listener)

    def add_trigger(self, trigger):
        self.__triggers.append(trigger)
        self.__triggers = sorted(self.__triggers)
        trigger.add_listener(self.on_trigger_executed)

    def execute(self, item_registry: ItemRegistry = ItemRegistry.instance()):
        for trigger in self.triggers:
            if isinstance(trigger, ManualTrigger):
                trigger.invoke(item_registry)
                return

    def on_trigger_executed(self, trigger: Trigger):
        for listener in self.__listeners:
            try:
                listener(self)
            except Exception as e:
                logging.warning("error occurred by calling listener", e)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __lt__(self, other):
        return self.__str__() < other.__str__()

    def __str__(self):
        return self.module + ".py#" + self.name + " (" + ", ".join(["'" + trigger.expression + "'" for trigger in self.__triggers]) + ")"

    def __repr__(self):
        return self.__str__()


class RuleEngine:

    __instance = None

    @staticmethod
    def instance():
        return RuleEngine.__instance

    @staticmethod
    def start_singleton(openhab_uri:str, python_rule_directory: str = "/etc/openhab/automation/rules", user: str = None, pwd: str = None):
        rule_engine = RuleEngine(openhab_uri, python_rule_directory, user, pwd)
        RuleEngine.__instance = rule_engine
        rule_engine.start()
        return rule_engine

    def __init__(self, openhab_uri:str, python_rule_directory: str, user: str, pwd: str):
        self.__listeners = set()
        self.python_rule_directory = python_rule_directory
        logging.info("connecting " + openhab_uri)
        ItemRegistry.new_singleton(openhab_uri, user, pwd)
        self.__event_consumer = EventConsumer(openhab_uri, self)
        self.__cron_scheduler = CronScheduler()
        self.__trigger_registry = TriggerRegistry()

    def add_listener(self, listener):
        self.__listeners.add(listener)

    def start(self):
        if self.python_rule_directory not in sys.path:
            sys.path.insert(0, self.python_rule_directory)
        for file in os.scandir(self.python_rule_directory):
            self.load_module(file.name)
        FileSystemListener.start(self)
        self.__cron_scheduler.start()
        self.__event_consumer.start()

    def add_trigger(self, trigger: Trigger):
        rules = self.rules
        if trigger.is_valid():
            logging.info(" * " + trigger.name + "(...): register trigger '" + trigger.expression + "'")
            self.__trigger_registry.register(trigger)
            if isinstance(trigger, CronTrigger):
                self.__cron_scheduler.add_job(trigger)
            elif isinstance(trigger, RuleLoadedTrigger):
                trigger.invoke(ItemRegistry.instance())
        else:
            logging.warning("Unsupported function spec " + trigger.module + "#" + trigger.name + " Ignoring it")
        if rules != self.rules:
            for listener in self.__listeners:
                try:
                    listener()
                except Exception as e:
                    logging.warning("error occured calling rules lsitener", e)

    def load_module(self, filename: str):
        if filename.endswith(".py"):
            try:
                modulename = self.__filename_to_modulename(filename)
                # reload?
                if modulename in sys.modules:
                    logging.info("reload '" + filename + "'")
                    self.__cron_scheduler.remove_jobs(modulename)
                    self.__trigger_registry.deregister(modulename)
                    importlib.reload(sys.modules[modulename])
                else:
                    logging.info("load '" + filename + "'")
                    importlib.import_module(modulename)
            except Exception as e:
                logging.warning("error occurred by (re)loading " + filename, e)

    def unload_module(self, filename: str):
        try:
            modulename = self.__filename_to_modulename(filename)
            if modulename in sys.modules:
                logging.info("unload '" + filename + "'")
                self.__cron_scheduler.remove_jobs(modulename)
                self.__trigger_registry.deregister(modulename)
                del sys.modules[modulename]
        except Exception as e:
            logging.warning("error occurred by unloading " + filename, e)

    def __filename_to_modulename(self, filename):
        return filename[:-3]

    def on_event(self, event):
        ItemRegistry.instance().on_event(event)
        triggers: List[ItemTrigger] = self.__trigger_registry.get_triggers_by_type(ItemTrigger)
        item_event = parse_item_event(event)
        if item_event is not None:
            matching_triggers = [trigger for trigger in triggers if trigger.matches(item_event)]
            for item_changed_trigger in matching_triggers:
                item_changed_trigger.invoke(ItemRegistry.instance())


    @property
    def rules(self) -> List[Rule]:
        rules = set()
        for module in self.__trigger_registry.triggers_by_module.keys():
            triggers = self.__trigger_registry.triggers_by_module[module]
            for func in [trigger.func for trigger in triggers]:
                rule = Rule(func)
                rule.add_trigger(ManualTrigger(func))
                for trigger in triggers:
                    if trigger.name == rule.name:
                        rule.add_trigger(trigger)
                rules.add(rule)
        return sorted(list(rules))



import logging
import time
import requests
import json
import sseclient
from threading import Thread
from typing import Optional
from dataclasses import dataclass


logging = logging.getLogger(__name__)




@dataclass()
class ItemEvent:
    item_name: str
    operation: str


def parse_item_event(event) -> Optional[ItemEvent]:
    topic = event.get("topic", "")
    if topic.startswith('openhab') or topic.startswith('smarthome'):
        try:
            parts = topic.split("/")
            if parts[1] == 'items':
                item_name = parts[2]
                operation = parts[3]
                return ItemEvent(item_name, operation)
        except Exception as e:
            logging.warning("Error occurred by handling event " + str(event), e)
    return None


class EventConsumer:

    def __init__(self, openhab_uri: str, event_listener):
        if not openhab_uri.endswith("/"):
            openhab_uri = openhab_uri + "/"
        self.event_uri =  openhab_uri + "rest/events"
        self.event_listener = event_listener
        self.is_running = True
        self.thread = None

    def start(self):
        self.thread = Thread(target=self.__listen, daemon=True)
        self.thread.start()

    def __listen(self):
        while self.is_running:
            try:
                logging.debug("opening sse stream (" + self.event_uri + ")")
                response = requests.get(self.event_uri, stream=True)
                client = sseclient.SSEClient(response)

                try:
                    for event in client.events():
                        data = json.loads(event.data)
                        self.event_listener.on_event(data)
                finally:
                    logging.debug("closing sse stream")
                    client.close()
                    response.close()
            except Exception as e:
                logging.error("error occurred consuming sse of " + self.event_uri, e)
                time.sleep(5)

    def stop(self):
        self.is_running = False
        Thread.join(self.thread)




import inspect
import logging
from logging import Logger


from openhab_pythonrule_engine.item_registry import ItemRegistry



logging = logging.getLogger(__name__)


class Invoker():

    TYPE_SINGLE_PARAM_ITEMREGISTRY = "TYPE_SINGLE_PARAM_ITEMREGISTRY"

    def __init__(self, func, type: str):
        self.__func = func
        self.name = func.__name__
        self.fullname = func.__module__ + "." + self.name
        self.__type = type

    @staticmethod
    def create(func):
        type = ""
        spec = inspect.getfullargspec(func)

        # one argument ItemRegistry
        if len(spec.args) == 1:
            type = Invoker.TYPE_SINGLE_PARAM_ITEMREGISTRY
            if spec.args[0] in spec.annotations:
                if spec.annotations[spec.args[0]] != ItemRegistry:
                    logging.warning("parameter " + str(spec.args[0]) + " is of type " + str(spec.annotations[spec.args[0]]) + ". " +
                                    str(spec.annotations[spec.args[0]]) + " is not supported (supported: ItemRegistry)")
                    return None
            else:
                logging.warning("assuming that parameter " + spec.args[0] + " is of type ItemRegistry. " \
                                "Please use type hints such as " + func.__name__ + "(" + spec.args[0]  + ": ItemRegistry)")
        return Invoker(func, type)

    def invoke(self, item_registry: ItemRegistry):
        if self.__type ==  self.TYPE_SINGLE_PARAM_ITEMREGISTRY:
            self.__func(item_registry)
        else:
            self.__func()
